//================================================================================
//
//  Window Desktop Window (DWM) / VSYNC / and other -- rapid prototype tests
//
//    - https://crbug.com/467617 (how to improve vsync in Chrome)
//    - https://crbug.com/422000 (Chrome 38 busts vsync)
//
//  AUTHOR:
//
//    Jerry Jongerius / jerryj@duckware.com
//
//  LICENSE:
//
//    - None.  This is my prototype code intended to move Chromium forward.  Chromium
//      may use this code in any way they see fit.  For all others, if you want to use
//      this code, just be nice and ask first...
//
//  HOW TO USE:
//
//    - don't run any other applications (confirm in task manager that cpu usage is zero)
//    - run "THIS.EXE >output.txt" from a DOS command prompt
//    - Tests / Run All Tests...
//    - exit app and review "output.txt"
//
//  PLATFORM:
//
//    - The intention is that this code will work on Vista/Win7/Win8/etc
//
//  WARNINGS:
//
//    - ALL of this code is a HACK -- ONLY rapid prototyping TO TEST OUT IDEAS AND SEE IF THEY WORK
//
//    - There sure seems to be a bunch of DWM bugs in Win7, so this code (or
//      similar) needs someone to extensively test under Vista
//
//    - review all 'TODO'
//
//  BUGBUG:
//
//    - DWM is crazy buggy! (especially on Win7)
//
//    - DwmFlush(): I run into a crazy multi-threading bug.  DwmFlush would return an
//      error code in one thread, but then later it would work.  It was 'fixed' only
//      after all threads actually called DwmFlush().  The bug only exists in
//      Win7 (not Win81), so MS must have known about it.
//
//    - DwmGetCompositionTimingInfo() info totally wrong!!  In multi-monitor testing, DwmGetCompositionTimingInfo
//      returned BOGUS vblank and period times.  The *period* returned was a constant 16.667ms, which was wrong.
//      and then the timings between vblanks was 16.667, 16.667, 26.xxx, (repeating over and over).  In Win7 I
//      can replicate by (1) starting in projector mode only, then (2) extending, then (3) switch primary back
//      to notebook display.  After that, DWM timings are corrupt.
//
//    - qpcRefreshPeriod from DwmGetCompositionTimingInfo() on Win7 may NOT change when another app is
//      maximized (and doing NO repaints).  And more bizarre is that qpcVBlank changes by this
//      unchanging/wrong qpcRefreshPeriod time.
//
//    - qpcRefreshPeriod from DwmGetCompositionTimingInfo() can be wrong!!
//      When we compute the period from two consecutive qpcVBlank from DwmGetCompositionTimingInfo
//      we don't always get the period reported by DwmGetCompositionTimingInfo.
//
//      Print out qpcRefreshPeriod and you will see a LOT of duplicates, as if there is some
//      throttle on how frequently qpcRefreshPeriod is computed?
//
//      On Win10, qpcRefreshPeriod can be outright wrong (16.666 constant)
//
//    - qpcVBlank on Win10 may not be prior vsync, but next vsync
//
//    - D3DKMTWaitForVerticalBlankEvent blocks DirectDraw's WaitForVerticalBlank
//      but only under Win7 (not Win81)
//
//    - DWM qpcVBlankFound accuracy: There is a strange situation on a Toshipa L775 laptop.  Ppressing
//      the FN key over and over (brings up/down a FN screen overlay) -- DWM qpcVBlank will be longer than
//      the normal (three frames over four frame period):
//        ... 16.712290, 16.716955, 19.383458, 22.718452, 24.764045, 16.719288, 16.718821, ...
//
//    - Win7: getting D3DKMT scan line while another thread is waiting for vsync causes problems
//
//
//  FUTURE RESEARCH:
//
//    - DXGK_INTERRUPT_CRTC_VSYNC - https://msdn.microsoft.com/en-us/library/ff559602(v=vs.85).aspx
//
//    - Multiple Monitors and Video Present Networks - https://msdn.microsoft.com/en-us/library/windows/hardware/ff568500
//
//  OBSERVATIONS:
//
//    - qpcRefreshPeriod does NOT always match manually calculating the difference between
//      two adjacent qpcVBlank.  This is subtle in Win7, and much more pronounced in Win8.
//
//    - It seems like a LOT of the minor jitter is actually due to the Windows OS and how
//      fast it can wake up a thread that is blocked on an event to trigger, which is why
//      test code was added to test the speed of the Windows message loop, and the speed
//      of waitForXXX.  Interesting results.
//
//    - Also, processor state has a huge impact on jitter.  Strangely, adding work
//      for the processor (preventing lower power states) can actually eliminate
//      a lot of jitter.
//
//================================================================================

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <ddraw.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>
#include "resource.h"
#include "backfill.h"

#define THREADPROC WINAPI

//======================================================================
//======================================================================
//======================================================================
//
//  Debug prints
//
//======================================================================
//======================================================================
//======================================================================

//----------------------------------------------------------------------
//void dd( LPSTR s ) {
//    printf( "%s\r\n", s );
//    fflush(stdout);
//    }
//----------------------------------------------------------------------
int tabs2spaces( LPSTR lpString ) {
    char buffer[1000];
    LPSTR in = lpString;
    LPSTR out = buffer;
    char c;
    while (c=*in++) {
        if (c=='\t') {
            while ((out-buffer)%12) {
                *out++ = ' ';
                }
            }
        else {
            *out++ = c;
            }
        }
    *out++ = '\0';
    CopyMemory( lpString, buffer, out-buffer );
    return (out-buffer)-(in-lpString);  // number of tabs added
    }
//----------------------------------------------------------------------
// output debug string
char *dout( char *format, ...) {
    
    // render debug string
    char temp[1000];
    va_list aptr;
    va_start(aptr, format);
    int j = vsprintf(temp, format, aptr);
    va_end(aptr);
    j += tabs2spaces( temp );
    j += sprintf(temp+j, "\r\n" );

    const int MAX=10*1024*1024;                  // This IS a hack - 10MB buffer
    static CRITICAL_SECTION dlock;
    static LPSTR lpBuffer=NULL;
    static LONG dpos=0;

    // self-init: MUST be called from WinMain (single-threaded) at least once before app goes multi-threaded
    if (!lpBuffer) {
        lpBuffer = (LPSTR)malloc(MAX);         // ... but this entire file is just prototype code, so OK
        memset( lpBuffer, '.', MAX );
        InitializeCriticalSectionAndSpinCount(&dlock,100);
        }

    // allocate space in global output buffer
    EnterCriticalSection(&dlock);
    int mypos = dpos;
    bool bAdd = mypos+j+1<MAX;
    dpos += (bAdd?j:0);
    lpBuffer[dpos] = '\0';
    LeaveCriticalSection(&dlock);
    if (bAdd) {
        CopyMemory( lpBuffer+mypos, temp, j );
        }

    return lpBuffer;
    }



//======================================================================
//======================================================================
//======================================================================

static LPSTR lpWindowClass="dwmvsynctests";
static HINSTANCE hAppInstance=NULL;
static HWND hAppWnd=NULL;
static HMENU hMainMenu=NULL;
static char szHEADING[500];
static char szPERCENTDONE[100];
static int nNumTestsRun=0;
static bool bAppRunning=true;
static bool bRunningTests=false;
static bool g_bSuspendVsyncServer=false;

static HBRUSH hRedBrush=NULL;
static HBRUSH hCyanBrush=NULL;
static int g_nFrame=0;

typedef __int64 OSTICK;   // an abtract unit of OS time (QPC value since init of this app)

void vsyncSuspend( bool );
LPSTR vsyncGetDebugMessage( LPSTR );
bool vsyncSleepWithVsyncInterrupt(DWORD);
bool vsyncGetNewTimingInfo( OSTICK*pvblank, OSTICK*pperiod );
void vsyncStartService();
void vsyncStopService();
OSTICK vsyncGetLast();
int vsyncGetScanLine();
int vsyncGSL();






//======================================================================
//======================================================================
//======================================================================
//
//  High res OSTICK for this program -- is the number of QPC ticks since program startup
//
//======================================================================
//======================================================================
//======================================================================

// NOTE: we don't care about math overflow conditions, since this is rapid
// prototyping code only meant to run for minutes

static __int64 _first;
static __int64 _freq;

//----------------------------------------------------------------------
// It is a curious note that Chrome tests to see if the processor supports
// invariant TSC before using Windows QPC, but then FAILS to verify that
// Windows QPC is then actually using invariant TSC as the time source!
// The QPC frequency is a dead giveaway.  There are well known values
// that indicate non-TSC -- that are MUCH slower than TSC.
void tickInit() {
    LARGE_INTEGER li;
    QueryPerformanceCounter( &li );
    _first = li.QuadPart;
    QueryPerformanceFrequency( &li );
    _freq = li.QuadPart;
    }
//----------------------------------------------------------------------
OSTICK qpc2tick( QPC_TIME qpc ) {
    return qpc-_first;
    }
//----------------------------------------------------------------------
OSTICK tick() {
    LARGE_INTEGER qpc;
    QueryPerformanceCounter( &qpc );
    return qpc2tick( *(QPC_TIME*)&qpc );
    }
//----------------------------------------------------------------------
double tick2us( OSTICK t ) {
    return t*1000000.0/_freq;
    }
//----------------------------------------------------------------------
double tick2ms( OSTICK t ) {
    return tick2us(t)/1000;
    }








//======================================================================
//======================================================================
//======================================================================

//
// The code below is a prototype written for Stanislav Chiknavaryan at
// chromium.org.  It shows how to very efficiently produce an accurate 
// timebase' and 'interval' from *ONLY* D3Dvsync wakeup times.
//
//   - See crbug.com/751340 - "Improve distribution of BeginFrame latency with D3DVsync experiment"
//   - See: http://www.vsynctester.com/firefoxisbroken.html#asolution
//   - See: http://www.vsynctester.com/howtocomputevsync.html
//

//----------------------------------------------------------------------
// hopefully to convey intent and assist in porting to Chromium, since
// this code must deal with nanosecond time (Chromium is just microseconds)
typedef double TimeTicks;
typedef double TimeDelta;
double TimeDelta2ms( TimeDelta t ) {
    return t; //t.ToInternalValue()/1000.0; // TODO PLACEHOLDER
    }
TimeDelta ms2TimeDelta( double ms ) {
    return ms; //TimeDelta::FromMillisecondsD(ms); // TODO PLACEHOLDER
    }

//----------------------------------------------------------------------
// vsync 'line of best fit' - see http://brownmath.com/stat/leastsq.htm for formulas
class LBF {

  private:

    // line of best fit vars
    double m_sx;            // sum x
    double m_sy;            // sum y
    double m_sxx;           // sum x*x
    double m_sxy;           // sum x*y
    size_t m_snum;          // sum number of points

    // other
    TimeTicks m_tFirst;     // base all incoming times from first time seen to maximize floating point precision
    double m_lastY;         // previous offset seen (for computing number of frames)
    size_t m_nFrame;        // computed 'X' value for algorithm

  public:

    //---
    LBF() {
         init();
         }

    //---
    void init() {
        m_sx = 0.0;
        m_sy = 0.0;
        m_sxx = 0.0;
        m_sxy = 0.0;
        m_snum = 0;
        m_tFirst = 0;
        m_lastY = 0.0;
        m_nFrame = 0;
        }

    //---
    void add( TimeTicks tNow, double interval ) {
        if (0==m_tFirst) {
            m_tFirst = tNow;
            }
        double Y = TimeDelta2ms(tNow-m_tFirst);
        m_nFrame += (size_t)((Y-m_lastY)/interval+0.5);
        m_lastY = Y;
        double X = m_nFrame;
        m_sx += X;
        m_sy += Y;
        m_sxx += X*X;
        m_sxy += X*Y;
        ++m_snum;
        }

    //---
    double slope() {
        return m_snum>2 ? (m_snum*m_sxy-m_sx*m_sy)/(m_snum*m_sxx-m_sx*m_sx) : 0;
        }
    };

//----------------------------------------------------------------------
class SCVSYNC {

  private:

    TimeTicks m_T0,m_T1,m_T2,m_T3,m_T4;  // last several times -- used to detect stable intervals (tight grouping)
    size_t m_nHits;              // number of tight grouping hits
    size_t m_nStable;            // stable interval (to microseconds) detection
    TimeTicks m_timebase;        // computed vsync timebase
    double m_interval;           // computed vsync interval (MUST *NOT* BE TIMEDELTA)
    LBF m_l1;                    // line of best fit
    LBF m_l2;                    // line of best fit

  public:

    //---
    SCVSYNC() {
        reset( "init", true );
        }

    //---
    void reset( LPSTR lpReason, bool bAll ) {
        //dout( "*****scReset(%s)", lpReason );
        if (bAll) {
            m_T4 = m_T3 = m_T2 = m_T1 = m_T0 = 0;
            }
        m_nHits = 0;
        m_nStable = 0;
        m_timebase = 0;
        m_interval = 0;
        m_l1.init();
        m_l2.init();
        }

    /*
    //---
    TimeTicks lowpoint() {
        TimeTicks ret = m_T0;
        double dip = 0;
        for (size_t loop=1; loop<=4; ++loop) {
            TimeTicks TL = 1==loop ? m_T1 : 2==loop ? m_T2 : 3==loop ? m_T3 : m_T4;
            double diff = TimeDelta2ms(TL-m_T0)-m_interval*loop;
            if (diff<dip) {
                dip = diff;
                ret = TL;
                }
            }
        return ret;
        }
    */

    //---
    TimeTicks haveNewTime( TimeTicks tNow, _Out_ TimeDelta*lpInterval ) {
        const double ONEMICROSECOND=.001;            // in millisecond units

        // error if wakeup times are not advancing (happens when DWM qpcVBlank is passed in and DWM does not advance) -- so DO NOTHING
        if (tNow<=m_T4) {
            return 0;
            }

        // shift in new time
        m_T0=m_T1; m_T1=m_T2; m_T2=m_T3; m_T3=m_T4; m_T4=tNow;

        // calculate grouping of all intervals
        TimeDelta D1 = m_T1-m_T0;
        TimeDelta D2 = m_T2-m_T1;
        TimeDelta D3 = m_T3-m_T2;
        TimeDelta D4 = m_T4-m_T3;
        TimeDelta vmin = min(D4,min(D3,min(D1,D2)));
        TimeDelta vmax = max(D4,max(D3,max(D1,D2)));
        double grouping = TimeDelta2ms(vmax-vmin);

        // if intervals are tightly grouped (within 500 microsecond for D3Dvsync), update interval statistics
        if (grouping<0.500) {
            ++m_nHits;
            double avems = TimeDelta2ms(m_T4-m_T0)/4;
            if (m_interval>0 && fabs(avems-m_interval)>1.0) { // detect change in Hz (interval changes >1ms)
                char msg[100];
                sprintf( msg, "Hz change %f to %f", m_interval, avems );
                reset( msg, false );
                }
            double msest = m_interval?m_interval:avems;
            m_l1.add( tNow, msest );
            m_l2.add( tNow, msest );
            if (0==(m_nHits%1000)) {
                m_l1 = m_l2;
                m_l2.init();
                }
            double ms = m_l1.slope();
            m_nStable = (ms>0 && fabs(ms-m_interval)<ONEMICROSECOND) ? m_nStable+1 : 0;
            m_interval = ms>0 ? ms : avems;

            // to be ultra-safe in timebase estimate, enable this
            //m_timebase = lowpoint();
            }

        // create vsync 'timebase' using negative and positive pressure bracketing
        double adjust=0;
        if (m_nStable>60 && m_timebase>0) {
            double ms = m_interval;
            adjust = (ms>0) ? fmod(fmod(TimeDelta2ms(tNow-m_timebase)+ms/8,ms)+ms,ms)-ms/8 : 0;
            }
        m_timebase = tNow - ms2TimeDelta(max(0,adjust)*0.999);

        // return 'timebase' and 'interval'
        *lpInterval = (m_interval>0) ? ms2TimeDelta(m_interval) : m_T3>0 ? (m_T4-m_T3) : ms2TimeDelta(1000.0/60);
        return m_timebase;
        }
    };








//======================================================================
//======================================================================
//======================================================================
//
//  Helper functions
//
//======================================================================
//======================================================================
//======================================================================



//----------------------------------------------------------------------
HANDLE newThread( LPTHREAD_START_ROUTINE lpfn ) {
    return CreateThread(NULL, 0, lpfn, NULL, 0, NULL);
    }
//----------------------------------------------------------------------
// we need to keep DWM busy so that vblank times are 'current'
// https://msdn.microsoft.com/en-us/library/ff569520(VS.85).aspx
void repaintApp() {
    if (hAppWnd) {
        InvalidateRect(hAppWnd, NULL, true);
        }
    }
//----------------------------------------------------------------------
void startTest( LPSTR lpName ) {
    static char SEP[]="====================================================================================================";
    static int EX=3;
    int slen = lstrlen(SEP);
    int w = EX+1+lstrlen(lpName)+1+EX;
    dout( "" );
    dout( SEP+slen-w );
    dout( "%s %s %s", SEP+slen-EX, lpName, SEP+slen-EX );
    dout( SEP+slen-w );
    dout( "" );
    lstrcpy( szHEADING, lpName );
    lstrcpy( szPERCENTDONE, "" );
    repaintApp();
    Sleep(10);
    }
//----------------------------------------------------------------------
boolean isTestStillRunning( int loop, int nIterations ) {
    sprintf( szPERCENTDONE, " -- on %d of %d (%.*f%%)", loop, nIterations, (nIterations>3000)?1:0, loop*100.0/nIterations );
    return bAppRunning && loop<nIterations;
    }

//----------------------------------------------------------------------
void spinwait( long us ) {
    OSTICK from = tick();
    while (tick2us(tick()-from)<us) {
        }
    }
//----------------------------------------------------------------------
//  This intent of this was to create a 'sleep' that can be vsync interrupted
//  because in Chrome, there must ultimately be a Sleep/WaitForXXX that we
//  want to convert to be vsync interrupted.
int sleepUntil( OSTICK to, bool bVsyncSleep ) {
    int nSpin=0;  // debug: how many times did we spin
    OSTICK now;
    while ((now=tick())<to) {
        ++nSpin;
        int ms = (int)ceil(tick2ms(to-now));
        if (bVsyncSleep) {
            bool bVsyncSignal = vsyncSleepWithVsyncInterrupt(ms>0?ms:1);
            if (bVsyncSignal) {
                break;
                }
            }
        else {
            Sleep( ms>0?ms:1 );
            }
        }
    return nSpin;
    }
//----------------------------------------------------------------------
void setWindowsTimeAccuracy( DWORD ms ) {
    timeBeginPeriod(ms);
    dout( "timeBeginPeriod: %d", ms );
    }


//======================================================================
//======================================================================
//======================================================================
//
//  Dynamic function binding
//
//======================================================================
//======================================================================
//======================================================================

//----------------------------------------------------------------------
FARPROC fnBind( LPSTR pDll, LPSTR lpFnName ) {
    HMODULE hMod = GetModuleHandle(pDll);
    if (!hMod) {
        hMod = LoadLibrary(pDll);
        }
    FARPROC lpfn = hMod ? GetProcAddress(hMod, lpFnName) : NULL;
    dout( "...fnBind(%s,%s)=%d", pDll, lpFnName, lpfn );
    //if (!lpfn) {
    //    dout( "*** Failed to bind to: %s:%s", pDll, lpFnName );
    //    }
    return lpfn;
    }
//----------------------------------------------------------------------
HRESULT myDwmFlush() {
    typedef HRESULT (WINAPI *LPFN)(void);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("dwmapi","DwmFlush");
        }
    return lpfn ? (*lpfn)() : E_NOTIMPL;
    }
//----------------------------------------------------------------------
HRESULT myDwmGetCompositionTimingInfo( HWND hWnd, DWM_TIMING_INFO *lpTI ) {
    typedef HRESULT (WINAPI *LPFN)(HWND, DWM_TIMING_INFO *);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("dwmapi","DwmGetCompositionTimingInfo");
        }
    return lpfn ? (*lpfn)(hWnd, lpTI) : E_NOTIMPL;
    }
//----------------------------------------------------------------------
// TODO/WARNING: The 'Note' at https://msdn.microsoft.com/en-us/library/windows/desktop/aa969518%28v=vs.85%29.aspx
// is curious, and basically says that DwmIsCompositionEnabled will incorrectly return FALSE some times
bool isDwmEnabled() {
    typedef HRESULT (WINAPI *LPFN)(BOOL*);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("dwmapi","DwmIsCompositionEnabled");
        }
    BOOL bEnabled;
    return lpfn ? (S_OK==(*lpfn)(&bEnabled)?0!=bEnabled:false) : false;
    }
//----------------------------------------------------------------------
BOOL myEnumDisplayDevices( LPCTSTR lpDevice, DWORD iDevNum, PDISPLAY_DEVICE lpDisplayDevice, DWORD dwFlags ) {
    typedef BOOL (WINAPI *LPFN)(LPCTSTR, DWORD, PDISPLAY_DEVICE, DWORD);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("user32","EnumDisplayDevicesA");
        }
    return lpfn ? (*lpfn)(lpDevice, iDevNum, lpDisplayDevice, dwFlags) : FALSE;
    }
//----------------------------------------------------------------------
NTSTATUS myD3DKMTWaitForVerticalBlankEvent( D3DKMT_WAITFORVERTICALBLANKEVENT*lpParams ) {
    typedef LPFND3DKMT_WAITFORVERTICALBLANKEVENT LPFN;
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("gdi32","D3DKMTWaitForVerticalBlankEvent");
        }
    return lpfn ? (*lpfn)(lpParams) : STATUS_NOT_IMPLEMENTED;
    }
//----------------------------------------------------------------------
NTSTATUS myD3DKMTOpenAdapterFromHdc( D3DKMT_OPENADAPTERFROMHDC*lpParams ) {
    typedef LPFND3DKMT_OPENADAPTERFROMHDC LPFN;
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("gdi32","D3DKMTOpenAdapterFromHdc");
        }
    return lpfn ? (*lpfn)(lpParams) : STATUS_NOT_IMPLEMENTED;
    }
//----------------------------------------------------------------------
NTSTATUS myD3DKMTCloseAdapter( D3DKMT_CLOSEADAPTER *lpParams ) {
    typedef LPFND3DKMT_CLOSEADAPTER LPFN;
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("gdi32","D3DKMTCloseAdapter");
        }
    return lpfn ? (*lpfn)(lpParams) : STATUS_NOT_IMPLEMENTED;
    }
//----------------------------------------------------------------------
NTSTATUS myD3DKMTGetScanLine( D3DKMT_GETSCANLINE *lpParams ) {
    typedef LPFND3DKMT_GETSCANLINE LPFN;
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("gdi32","D3DKMTGetScanLine");
        }
    return lpfn ? (*lpfn)(lpParams) : STATUS_NOT_IMPLEMENTED;
    }

//----------------------------------------------------------------------
NTSTATUS myD3DKMTQueryAdapterInfo( D3DKMT_QUERYADAPTERINFO *lpQuery ) {
    typedef LPFND3DKMT_QUERYADAPTERINFO LPFN;
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("gdi32","D3DKMTQueryAdapterInfo");
        }
    return lpfn ? (*lpfn)(lpQuery) : STATUS_NOT_IMPLEMENTED;
    }


//----------------------------------------------------------------------
bool getCurrentDisplayMode( D3DKMT_HANDLE hAdapter, D3DDDI_VIDEO_PRESENT_SOURCE_ID VidPnSourceId, D3DKMT_DISPLAYMODE *lpMode ) {
    D3DKMT_CURRENTDISPLAYMODE dm;
    dm.VidPnSourceId = VidPnSourceId;
    D3DKMT_QUERYADAPTERINFO ai;
    ai.hAdapter = hAdapter;
    ai.Type = KMTQAITYPE_CURRENTDISPLAYMODE;
    ai.pPrivateDriverData = &dm;
    ai.PrivateDriverDataSize = sizeof(dm);
    if (STATUS_SUCCESS==myD3DKMTQueryAdapterInfo(&ai)) {
        *lpMode = dm.DisplayMode;
        return true;
        }
    return false;
    }










//----------------------------------------------------------------------
HMONITOR myMonitorFromWindow( HWND hWnd, DWORD dwFlags ) {
    typedef HMONITOR (WINAPI *LPFN)(HWND, DWORD);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("user32","MonitorFromWindow");
        }
    return lpfn ? (*lpfn)(hWnd, dwFlags) : NULL;
    }
//----------------------------------------------------------------------
BOOL myGetMonitorInfo( HMONITOR hMonitor, LPMONITORINFOEX lpmi) {
    typedef BOOL (WINAPI *LPFN)(HMONITOR, LPMONITORINFOEX);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("user32","GetMonitorInfoA");
        }
    return lpfn ? (*lpfn)(hMonitor, lpmi) : FALSE;
    }

//----------------------------------------------------------------------
NTSTATUS myNtQueryTimerResolution(__int64*pmaxms, __int64*pminms, __int64*pcurms) {
    typedef NTSTATUS (WINAPI *LPFN)(__int64*, __int64*, __int64*);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("ntdll","NtQueryTimerResolution");
        }
    return lpfn ? (*lpfn)(pmaxms, pminms, pcurms) : STATUS_NOT_IMPLEMENTED;
    }
//----------------------------------------------------------------------
HRESULT myDirectDrawCreate(GUID FAR* lpGUID, LPDIRECTDRAW FAR* lplpDD, IUnknown FAR* pUnkOuter) {
    typedef HRESULT (WINAPI *LPFN)(GUID FAR* lpGUID, LPDIRECTDRAW FAR* lplpDD, IUnknown FAR* pUnkOuter);
    static bool bNeedInit=true;
    static LPFN lpfn;
    if (bNeedInit) {
        bNeedInit = false;
        lpfn = (LPFN)fnBind("ddraw","DirectDrawCreate");
        }
    return lpfn ? (*lpfn)(lpGUID, lplpDD, pUnkOuter) : E_NOTIMPL;
    }


           
//======================================================================
//======================================================================
//======================================================================
//
//  DWM / VSYNC / etc Tests
//
//======================================================================
//======================================================================
//======================================================================

bool getDwmTimingInfo( _Out_Opt_ OSTICK*pvblank, _Out_Opt_ OSTICK*pperiod, UNSIGNED_RATIO*pp2, _In_ bool bLogError ) {
    bool bRet=false;
    DWM_TIMING_INFO TI;
    TI.cbSize = sizeof(TI);
    HRESULT res = myDwmGetCompositionTimingInfo(NULL, &TI);   // NOTE: docs say hWnd *MUST* be NULL in Win81, or E_INVALIDARG will be returned
    if (S_OK==res) {
        if (pvblank) {
            *pvblank = qpc2tick(TI.qpcVBlank);
            }
        if (pperiod) {
            *pperiod = TI.qpcRefreshPeriod;
            }
        if (pp2) {
            *pp2 = TI.rateRefresh;
            }
        bRet = true;
        }
    else if (bLogError) {
        dout( "GTI ERROR %d", res );
        }

    return bRet;
    }

//----------------------------------------------------------------------
// microsecond time offset from DWM's qpcVBlank (presumed to be 'true vsync').
// This code assumes that it will only really be called once per vsync iteration
// by other test code to self monitor accuracy compared to DWM timing
// WARNING: On inactivity, DWM stops updating timing info
double dwmoff( _In_ OSTICK toff ) {
    static OSTICK g_vlast=0;

    OSTICK vblank;
    OSTICK vperiod;
    bool bOK=false;

    // this crazyness is needed because sometimes DWM timing info is BEHIND us (we are running high priority) and we need to give DWM CPU cycles to catch up
    HANDLE hThread = GetCurrentThread();
    int oldprio = GetThreadPriority( hThread );
    SetThreadPriority( hThread, THREAD_PRIORITY_NORMAL );
    OSTICK t1 = tick();
    // wait up to 0.5ms for DWM to catch up
    while ((bOK=getDwmTimingInfo(&vblank,&vperiod,NULL,false)) && vblank==g_vlast && tick2us(tick()-t1)<500) {
        Sleep(0); // Yield
        }
    SetThreadPriority( hThread, oldprio );

    double ret = bOK ? tick2us(toff-vblank) : 0;
    g_vlast = vblank;
    return ret;
    }

//----------------------------------------------------------------------
// https://msdn.microsoft.com/en-us/library/aa911354.aspx (confirms WaitForVerticalBlank is cpu hog)
void testDirectDraw( int nIterations, bool bExtended ) {
    startTest( "DirectDraw WaitForVerticalBlank" );

    LPDIRECTDRAW pDDraw=NULL;
    if (DD_OK==myDirectDrawCreate(NULL, &pDDraw, NULL) && pDDraw) {  // NULL first arg causes DD to use primary adapter
        dout( "iteration\tvsync/qpc\tperiod/ms\tinvblank/ms\tvoff/us" );
        OSTICK last=0;
        for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
            // BUGBUG: On Win7 this blocks if D3DKMTWaitForVerticalBlankEvent gets there first; But no bug on Win8
            pDDraw->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, NULL);
            OSTICK in = tick();
            pDDraw->WaitForVerticalBlank(DDWAITVB_BLOCKEND, NULL);
            OSTICK out = tick();
            if (last) {
                dout( "%d\t%I64d\t%.3f\t%.3f\t%.1f", loop, in, tick2ms(in-last), tick2ms(out-in), dwmoff(in) );
                }
            last = in;
            if (!bExtended) {
                repaintApp();
                }
            }
        }
    }

//----------------------------------------------------------------------
// refresh rate -- added to show Mason Chang <mchang@mozilla.com> that Firefox is using the wrong 'rate'
void testFirefoxRateRefresh() {
    startTest( "Demonstrate that Firefox is using the wrong display Hz" );

    //
    //  Use on computer where the true Hz is NOT near an integer rounded value
    //
    if (true) {
        OSTICK vperiod;
        UNSIGNED_RATIO rr;;
        if (getDwmTimingInfo(NULL,&vperiod,&rr,true)) {
            // Firefox computation (from gfxWindowsPlatform.cpp)
            float rate = ((float)rr.uiDenominator / (float)rr.uiNumerator) * 1000;
            dout( "dwm.rateRefresh = %.6f/ms from %u/%u*1000 (used by Firefox and NOT accurate)", rate, rr.uiDenominator, rr.uiNumerator );
            dout( "dwm.qpcRefreshPeriod = %.6f/ms (more accurate on Win7/8 but not good on Win10)", tick2ms(vperiod) );
            }
        }

    
    //
    //   SEE NEW 'SCVSYNC' CODE
    //
    /*
    dout( "" );
    dout( "iteration\tvblank/qpc\tperiod/ms\tave/ms\tbest/ms" );
    double sx=0.0;
    double sy=0.0;
    double sxx=0.0;
    double sxy=0.0;
    double sum=0;
    int nsum=0;
    double msest = 16.66;
    OSTICK prev=0;
    OSTICK vblank;
    int nFound=0;
    while (isTestStillRunning(nFound,60*20) && getDwmTimingInfo(&vblank,NULL,NULL,true)) {
        //todo: set msest on dwm period
        if (vblank!=prev) {
            if (prev!=0) {
                //todo: there needs to be a filter to only use 'good' vsync times from DWM
                double diff = tick2ms(vblank-prev);
                int nFrames = (int)(diff/msest+0.5);
                // todo: (1) update msest
                // todo: (2) reset if found ms differs from msest by too much (frequency has changed / etc)
                sum += diff;
                nsum += nFrames;
                double X = nsum;
                double Y = tick2ms(vblank);
                sx += X;
                sy += Y;
                sxx += X*X;
                sxy += X*Y;
                ++nFound;
                dout( "%d\t%I64d\t%.6f\t%.6f\t%.6f", nFound, vblank, diff, sum/nsum, (nFound*sxy-sx*sy)/(nFound*sxx-sx*sx) );
                }
            prev = vblank;
            repaintApp();
            }
        Sleep(0);
        }
    */
    }

//----------------------------------------------------------------------
// https://msdn.microsoft.com/en-us/library/windows/desktop/dd743679(v=vs.85).aspx
void playWaveResource( LPCSTR lpName ) {
    HRSRC hResInfo = FindResource(hAppInstance, lpName, "WAVE");
    HGLOBAL hSound = (NULL!=hResInfo) ? LoadResource(hAppInstance, hResInfo) : NULL;
    LPCSTR lpSound = (NULL!=hSound) ? (LPCSTR)LockResource(hSound) : NULL;
    if (NULL!=lpSound) {
        sndPlaySound( lpSound, SND_MEMORY|SND_ASYNC|SND_NODEFAULT );
        }
    }

//----------------------------------------------------------------------
double fabs(double v) {
    return v<0 ? -v : v;
    }
//----------------------------------------------------------------------
// SPIN WAIT for vblank from dwm to change (WILL NOT CHANGE UNTIL SCREEN IS DOING SOME PAINTING)
void testDwmGetCompositionTimingInfo( int nIterations, bool bExtended ) {
    startTest( "DwmGetCompositionTimingInfo" );
    OSTICK prev=0;
    OSTICK vblank;
    OSTICK period;
    int nFound=0;

    double pms=0;

    dout( "iteration\tnow/qpc\tvblank/qpc\tdiff/us\tperiod/qpc\tperiod/ms\tdelta/ms" );
    while (isTestStillRunning(nFound,nIterations) && getDwmTimingInfo(&vblank,&period,NULL,true)) {
        if (vblank!=prev) {
            OSTICK now = tick();
            double ms = tick2ms(prev?vblank-prev:0);
            dout( "%d\t%I64d\t%I64d\t%.0f\t%I64d\t%.3f\t%.3f", nFound, now, vblank, tick2us(now-vblank), period, tick2ms(period), ms );
            prev = vblank;
            ++nFound;
            repaintApp();

            // Show Firefox bug time bug: beep when there is a REAL DWM inter-frame difference greater than 500 microseconds
            if (bExtended && (pms>0) && (ms>0) && (fabs(pms-ms)>0.500)) {
                //MessageBeep(MB_ICONHAND);
                playWaveResource("wavbeep");
                dout( "--> %f %f", pms, ms );
                }
            pms = ms;
            }
        Sleep(0);
        }
    }

//----------------------------------------------------------------------
void testDwmFlush( int nIterations ) {
    startTest( "DwmFlush" );
    dout( "iteration\tnow/qpc\tdwm/qpc\tperiod/ms\tvoff/us\tffvbl/qpc\tffvsc/qpc" );
    OSTICK last=0;
    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
        HRESULT res = myDwmFlush();
        if (S_OK==res) {
           OSTICK now = tick();
           if (last) {
                OSTICK vblank;
                OSTICK vperiod;
                if (getDwmTimingInfo(&vblank,&vperiod,NULL,true)) {
                    // added code to show Mason Chang <mchang@mozilla.com> / Firefox how to conform 'now' to 'vsync'
                    OSTICK ffvblank = vblank + (loop%5-2)*vperiod; // FF: mimic DWM that is behind/current/ahead
                    OSTICK ffvsync = now - (((now-ffvblank+vperiod/8)%vperiod+vperiod)%vperiod-vperiod/8);  // FF: show how to conform 'now' to 'prior vsync'
                    dout( "%d\t%I64d\t%I64d\t%.3f\t%.1f\t%I64d\t%I64d", loop, now, vblank, tick2ms(now-last), dwmoff(now), ffvblank, ffvsync );
                    }
                }
            last = now;
            repaintApp();
            }
        }
    }

//----------------------------------------------------------------------
void testScanLine( int nIterations ) {
    startTest( "ScanLine" );
    dout( "iteration\tscanline" );

    if (true) {
        OSTICK t1 = tick();
        for (int loop=0; loop<1000; ++loop) {
            vsyncGetScanLine();
            }
        OSTICK t2 = tick();
        dout( "vsyncGetScanLine() runs in %.1f us", tick2us(t2-t1)/1000 );
        }

    for (int loop=0; isTestStillRunning(loop,1200); ++loop) {
        UINT scanline = vsyncGetScanLine();
        dout( "%d\t%d", loop, scanline );
        spinwait(15);
        }
    }


//----------------------------------------------------------------------
void testMimicAnimationLoop( int nIterations ) {
    startTest( "Mimic animation (waiting for vsync intervals)" );

    dout( "want/qpc\tgot/qpc\tlate/qpc\tperiod/ms\tspin/int\tvoff/us" );
    OSTICK last=0;
    OSTICK vNext=0;
    bool bUseNew = false;

    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
        OSTICK vsync;
        OSTICK period;

        if (vsyncGetNewTimingInfo(&vsync,&period)) {
            // advance to NEXT vsync
            while (vsync<tick() || vsync<=vNext) {
                vsync += period;
                };
            bool bWantUseNew = loop>nIterations/4 && loop<nIterations*3/4;
            if (bWantUseNew!=bUseNew) {
                bUseNew = bWantUseNew;
                dout( "******** %s", bUseNew?"ON":"OFF" );
                }
            int nSpin = sleepUntil(vsync, bUseNew);     
            OSTICK now = tick();
            if (last) {
                dout( "%I64d\t%I64d\t%I64d\t%.3f\t%d\t%.1f", vsync, now, (now-vsync), tick2ms(now-last), nSpin, dwmoff(now) );
                }
            last = now;
            repaintApp();
            }
        else {
            dout( "timing info is not yet available..." );
            Sleep(20);
            }
        }
    }

//----------------------------------------------------------------------
void giveDwmTimeToFinish() {
    OSTICK t1=tick();
    while (tick2ms(tick()-t1)<1) {
        Sleep(0);  // try to give DWM time to run finish updating stats
        }
    }
//----------------------------------------------------------------------
void testStanislavChromium( int nIterations ) {
    startTest( "testStanislavChromium" );

    dout( "iteration\tD3Dnow/ms\tdelta/ms\tSCvsync/ms\tinterval/ms\tDWMnow/ms\tDWMvsync/ms\tinterval/ms\toverhead/us\tscanline" );

    SCVSYNC scvsync;
    SCVSYNC scdwm;

    OSTICK nowlast=0;
    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {

        // D3DVsync stats
        vsyncSleepWithVsyncInterrupt(100);
        OSTICK now = vsyncGetLast();

        double tNow = tick2ms(now);
        double interval=0;
        //dout( "scnow %I64d", now );
        OSTICK t1 = tick();
        double vsync = scvsync.haveNewTime(tNow, &interval);
        OSTICK t2 = tick();

        // DWM stats
        double dwmraw=0;
        double dwmvsync=0;
        double dwminterval=0;
        OSTICK dwm=0;
        giveDwmTimeToFinish();
        if (getDwmTimingInfo(&dwm,NULL,NULL,true)) {
            if (dwm) {
                dwmraw = tick2ms(dwm);
                //dout( "dwmnow %I64d", dwm );
                dwmvsync = scdwm.haveNewTime(dwmraw, &dwminterval);
                }
            }

        // check for D3Dvsync and DWM 'out of sync'
        if (interval>0) {
            double x = fmod(fabs(tNow-dwmraw),interval);
            if (x>2.0 && (interval-x)>2.0) {
                dout( "XXXXX %f %f", tNow, dwmraw );
                playWaveResource("wavbeep");
                }
            }

        int scanline = vsyncGSL();

        
        if ((loop<10000) || (0==loop%10)) {
            dout( "%d\t%.4f\t%.3f\t%.4f\t%.6f\t%.4f\t%.4f\t%.6f\t%.3f\t%d", loop, tick2ms(now), (nowlast?tick2ms(now-nowlast):0), vsync, interval, dwmraw, dwmvsync, dwminterval, tick2us(t2-t1), scanline );
            }
        nowlast = now;
        repaintApp();
        }
    }

//----------------------------------------------------------------------
void testWaitForBlanking( int nIterations ) {
    startTest( "testWaitForBlanking" );

    dout( "now/qpc\tperiod/ms\tdwm/qpc\tperiod/ms" );
    OSTICK nowlast=0;
    OSTICK dwmlast=0;

    // test for how well myD3DKMTWaitForVerticalBlankEvent works relative to DWM functions

    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
        vsyncSleepWithVsyncInterrupt(50);
        OSTICK now = vsyncGetLast(); //tick();
        OSTICK dwm=0;
        getDwmTimingInfo(&dwm,NULL,NULL,true);
        if (nowlast) {
            dout( "%I64d\t%.3f\t%I64d\t%.3f", now, tick2ms(now-nowlast), dwm, tick2ms(dwm-dwmlast) );
            }
        nowlast = now;
        dwmlast = dwm;
        repaintApp();
        }
    }



//----------------------------------------------------------------------
void doSleepTest( DWORD ms ) {
    for (int loop=0; loop<20; ++loop) {
        Sleep(1);             // sync to OS time quantum boundary
        spinwait(loop*1000);  // offset so many milliseconds
        OSTICK t1 = tick();
        Sleep(ms);
        OSTICK t2 = tick();
        double tms = tick2ms(t2-t1);
        double err = tms-ms;
        dout( "%d\t%.3f\t%.3f\t%.3f", loop, ms*1.0, tms, err );
        }
    }
//----------------------------------------------------------------------
void testWindowsSleepAccuracy() {
    startTest( "OS Sleep Accuracy, at various quantum offsets" );
    dout( "offset/ms\tsleep/ms\tslept/ms\terror/ms" );
    doSleepTest(0);
    doSleepTest(1);
    doSleepTest(10);
    }

//----------------------------------------------------------------------
void testQpcSpeed() {
    startTest( "QPC Speed" );
    tick();  // hit once before timing

    long lNum=0;
    OSTICK t1 = tick();
    OSTICK t2 = t1;
    while (tick2ms(t2-t1)<1000) {
        for (int loop=0; loop<10000; ++loop) {
            tick(); tick(); tick();
            tick(); tick(); tick();
            tick(); tick(); tick();
            t2 = tick();
            lNum += 10;
            }
        }
    dout( "%.0f ns per tick() call (%d calls in %I64d/qpc)", tick2us(t2-t1)*1000/lNum, lNum, (t2-t1) );
    }


//----------------------------------------------------------------------
void testWindowsTimeQuantum() {
    startTest( "Windows Time Quantum" );

    // ask Windows kernal for settings
    __int64 imaxms=0;
    __int64 iminms=0;
    __int64 icurms=0;
    if (STATUS_SUCCESS==myNtQueryTimerResolution(&imaxms, &iminms, &icurms)) {
        dout( "Windows Kernel reports: min/ms=%f  max/ms=%f   cur/ms=%f", iminms/10000.0, imaxms/10000.0, icurms/10000.0 );
        dout( "" );
        }

    // then try to measure setting
    dout( "iteration\tquantum/ms" );
    for (int loop=0; loop<10; ++loop) {
        Sleep(1);
        OSTICK t1 = tick();
        Sleep(1);
        OSTICK t2 = tick();
        dout( "%d\t%.3f", loop, tick2ms(t2-t1) );
        }
    }

//----------------------------------------------------------------------
// https://msdn.microsoft.com/en-us/library/windows/hardware/ff568606(v=vs.85).aspx "OpenGL Functions Called by OpenGL Installable Client Drivers"
// https://msdn.microsoft.com/en-us/library/windows/hardware/ff569520%28v=vs.85%29.aspx "Saving Energy with VSync Control"
// works even with DWM is NOT working
void testD3DKMT( int nIterations, bool bExtended ) {
    startTest( "D3DKMTWaitForVerticalBlankEvent" );

    DISPLAY_DEVICE dd;
    dd.cb = sizeof(dd);
    for (int loop=0; myEnumDisplayDevices(NULL, loop, &dd, 0); ++loop) {
        if (dd.StateFlags&DISPLAY_DEVICE_ACTIVE) {
            HDC hDC = CreateDC(NULL, (LPSTR)dd.DeviceName, NULL, NULL);
            dout( "--> %s (%s) hDC=%d", dd.DeviceName, dd.DeviceString, hDC );
            if (hDC) {
                D3DKMT_OPENADAPTERFROMHDC oa;
                oa.hDc = hDC;
                NTSTATUS status = myD3DKMTOpenAdapterFromHdc(&oa);
                dout( "oa=%d", status );  // want 0
                DeleteDC( hDC );
                D3DKMT_WAITFORVERTICALBLANKEVENT we;
                we.hAdapter = oa.hAdapter;
                we.hDevice = 0;
                we.VidPnSourceId = oa.VidPnSourceId;
                //status = (*lpfnKTWaitForVerticalBlankEvent)(&we);
                //dout( "we=%d", status );  // want 0

                dout( "iteration\tvsync/qpc\tperiod/ms\tvoff/us" );
                OSTICK last=0;
                for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
                    status = myD3DKMTWaitForVerticalBlankEvent(&we);
                    OSTICK now = tick();
                    if (STATUS_SUCCESS==status) {
                        if (last) {
                            dout( "%d\t%I64d\t%.3f\t%.1f", loop, now, tick2ms(now-last), dwmoff(now) );
                            }
                        last = now;
                        if (!bExtended) {
                            repaintApp();
                            }
                        }
                    else {
                        dout( "err %d", status );
                        Sleep(100);
                        }
                    }
                }
            }
        }
    }


//----------------------------------------------------------------------
void TestWindowsMessageSpeed( int nIterations ) {
    startTest( "Windows Message Loop Processing Speed" );
    dout( "iteration\tdelay/us" );
    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
        Sleep(20);
        PostMessage( hAppWnd, WM_USER+7, (WPARAM)loop, (LPARAM)tick() );
        }
    Sleep(100);
    }




//
// test Widnows WaitForXXX speed
//

HANDLE hPingPong;
OSTICK tPingPong;
//----------------------------------------------------------------------
DWORD THREADPROC runPong( LPVOID lpParam ) {
    HANDLE hWait;
    while (hWait=hPingPong) {
        OSTICK tBegin=tick();
        if (WAIT_OBJECT_0==WaitForSingleObject(hWait, 100)) {
            OSTICK t2 = tick();
            if ((t2-tBegin)>30) {
                dout( "%.3f", tick2us(t2-tPingPong) );
                }
            }
        else {
            break;
            }
        }
    return 0;
    }
//----------------------------------------------------------------------
void testWindowsWaitForXXXSpeed( int nIterations ) {
    timeBeginPeriod(1);
    startTest( "Windows WaitForXXXX Processing Speed" );
    dout( "delay/us" );
    hPingPong = CreateEvent(NULL, FALSE, FALSE, NULL);
    newThread( runPong );
    for (int loop=0; isTestStillRunning(loop,nIterations); ++loop) {
        Sleep(50);
        tPingPong = tick();
        SetEvent( hPingPong );
        }
    hPingPong = NULL;
    timeEndPeriod(1);
    Sleep(100);
    }

//
//  Output displays
//

//----------------------------------------------------------------------
// https://msdn.microsoft.com/en-us/library/windows/hardware/ff547047%28v=vs.85%29.aspx
void outputDisplays() {
    startTest( "Display Adapter Information" );
    DISPLAY_DEVICE dd;
    dd.cb = sizeof(dd);
    for (int loop=0; myEnumDisplayDevices(NULL, loop, &dd, 0); ++loop) {
        bool bPrimary = (dd.StateFlags&DISPLAY_DEVICE_PRIMARY_DEVICE)!=0;
        bool bActive = (dd.StateFlags&DISPLAY_DEVICE_ACTIVE)!=0;
        dout( "Adaptor%d=%s '%s' %s%s", loop, dd.DeviceName, dd.DeviceString, (bActive?" [ACTIVE]":""), (bPrimary?" [PRIMARY]":"") );
        //DISPLAY_DEVICE md;
        //md.cb = sizeof(md);
        //for (int loop2=0; myEnumDisplayDevices((LPSTR)dd.DeviceName, loop2, &md, 0); ++loop2) {
        //    dout( " Monitor=%s '%s'", md.DeviceName, md.DeviceString );
        //    }
        }
    }


//
//  Find out what monitor this test application is (mostly) running on
//

//----------------------------------------------------------------------
void outputAppOnMonitorInfo() {
    HMONITOR hMonitor = myMonitorFromWindow(hAppWnd, MONITOR_DEFAULTTONEAREST);
    MONITORINFOEX mi;
    mi.cbSize = sizeof(mi);
    if (hMonitor && myGetMonitorInfo(hMonitor, &mi)) {
        bool bPrimary = (mi.dwFlags&MONITORINFOF_PRIMARY)!=0;

        /*
        int width=0;
        int height=0;
        HDC hDC = CreateDC(NULL, mi.szDevice, NULL, NULL);
        if (hDC) {
            //width = GetDeviceCaps(hDC, HORZRES);
            //height = GetDeviceCaps(hDC, VERTRES);
            width = GetSystemMetrics (SM_CXSCREEN); 
            height = GetSystemMetrics (SM_CYSCREEN);
            DeleteDC( hDC );
            }
        */

        dout( "" );
        dout( "App running on %s%s", mi.szDevice, (bPrimary?" (PRIMARY)":"") );
        }
    }


//
//
//

//----------------------------------------------------------------------
DWORD THREADPROC runAllTestsThread( LPVOID lpParam ) {
    bRunningTests = true;
    dout( "+runAllTests()" );

    vsyncSuspend(true);
    SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_HIGHEST );
    outputDisplays();
    outputAppOnMonitorInfo();

    
    testQpcSpeed();
    testWindowsTimeQuantum();
    testWindowsSleepAccuracy();
    TestWindowsMessageSpeed(10);
    testWindowsWaitForXXXSpeed(10);

    int nIterations=60;
    testScanLine(nIterations);
    testDirectDraw(nIterations, false);
    testD3DKMT(nIterations, false);
    testFirefoxRateRefresh();
    testDwmGetCompositionTimingInfo(nIterations*10,false);
    testDwmFlush(nIterations);

    vsyncSuspend(false);

    testMimicAnimationLoop(nIterations);
    testWaitForBlanking(600);
    testStanislavChromium(1000);

    startTest( "*** To view test results, exit application ***" );
    Sleep(3000);
    //startTest( "DONE" );

    dout( "-runAllTests()" );
    bRunningTests = false;

    MessageBeep(0);
    return 0;
    }
//----------------------------------------------------------------------
DWORD THREADPROC runExtendedDirectDraw( LPVOID lpParam ) {
    bRunningTests = true;
    vsyncSuspend(true);
    testDirectDraw(600, true);
    vsyncSuspend(false);
    bRunningTests = false;

    MessageBeep(0);
    return 0;
    }
//----------------------------------------------------------------------
DWORD THREADPROC runExtendedD3DKMT( LPVOID lpParam ) {
    bRunningTests = true;
    vsyncSuspend(true);
    testD3DKMT(600, true);
 

//    // TODO: testing DwmFlush during screen off
//    for (int x=0; x<2; ++x) {
//        Sleep(30000);
//        for (int loop=0; loop<10; ++loop) {
//            HRESULT res = myDwmFlush();
//            if (S_OK==res) {
//                OSTICK now = tick();
//                dout( "%d\t%I64d", loop, now );
//                }
//            }
//        }

    
    vsyncSuspend(false);
    bRunningTests = false;

    MessageBeep(0);
    return 0;
    }

//----------------------------------------------------------------------
DWORD THREADPROC runExtendedCompositionTimingInfo( LPVOID lpParam ) {
    bRunningTests = true;
    vsyncSuspend(true);
    testDwmGetCompositionTimingInfo(3600,true);
    vsyncSuspend(false);
    bRunningTests = false;

    MessageBeep(0);
    return 0;
    }

//----------------------------------------------------------------------
DWORD THREADPROC runExtendedStanislavChromiumTest( LPVOID lpParam ) {
    bRunningTests = true;
    SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_HIGHEST );
    testStanislavChromium(100000);
    bRunningTests = false;
    MessageBeep(0);
    return 0;
    }








int _m1x=0;      // last mouse X
int _m1y=0;      // last mouse Y
OSTICK _m1t=0;   // last mouse tick

int _m2x=0;      // last mouse X
int _m2y=0;      // last mouse Y
OSTICK _m2t=0;   // last mouse tick

//----------------------------------------------------------------------
void doPaintWindow( HWND hWnd ) {
    static int nNumPaints=0;


    PAINTSTRUCT ps;
    HDC hdc = BeginPaint(hWnd, &ps);

    RECT r;

    //r.left = 0;
    //r.top = 200;
    //r.bottom = r.top+200;
    //r.right = r.left+200;
    //FillRect( hdc, &r, g_nFrame&1 ? hRedBrush : hCyanBrush );

    GetClientRect( hWnd, &r );
    char msg[1000];
    char vdebug[200];
    char headline[200];
    sprintf( headline, "%s%s", szHEADING, szPERCENTDONE );
    sprintf( msg, "%s\n\n%s\nnNumPaints=%d    MOUSE x=%d y=%d ms=%.3f", vsyncGetDebugMessage(vdebug), (bRunningTests?headline:""), ++nNumPaints, _m1x, _m1y, tick2ms(_m1t-_m2t) );
    DrawText( hdc, msg, lstrlen(msg), &r, DT_CENTER );

    // box around cursor pos
    r.left = _m1x-4;
    r.right = _m1x+4;
    r.top = _m1y-4;
    r.bottom = _m1y+4;
    //FillRect( hdc, &r, hRedBrush ); //GetSysColorBrush(COLOR_WINDOWTEXT) );
    //FrameRect( hdc, &r, hRedBrush );
    FillRect( hdc, &r, hRedBrush );



    EndPaint( hWnd, &ps );
    }


//======================================================================
//======================================================================
//======================================================================
//
//  The generated 'Hello World' app
//
//======================================================================
//======================================================================
//======================================================================

//----------------------------------------------------------------------
// TESTING HACK.  Place this app window and a browser window to "http://www.vsynctester.com/testing/mouse.html"
// side by side, and then couble click in this app window (to see/compare input lag)
DWORD THREADPROC slideCursorThread(LPVOID lpParam) {
    if (!bRunningTests) {
        dout( "+mouseslide" );
        for (int loop=0; isTestStillRunning(loop,60); ++loop) {
            vsyncSleepWithVsyncInterrupt(33);
            POINT pt;
            if (GetCursorPos(&pt)) {
                SetCursorPos(pt.x+20, pt.y);
                }
            }
        dout( "-mouseslide" );
        }
    return 0;
    }

//----------------------------------------------------------------------
BOOL isRepaintOnMouseMove() {
    return MF_CHECKED&GetMenuState(hMainMenu, IDM_SETTINGS_REPAINTONMOUSEMOVE, 0);
    }

//----------------------------------------------------------------------
BOOL CALLBACK AboutDialogBox( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam ) {
    switch (message) {
		case WM_INITDIALOG:
            return TRUE;
		case WM_COMMAND:
            if (LOWORD(wParam)==IDOK || LOWORD(wParam)==IDCANCEL) {
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
                }
			break;
        }
    return FALSE;
    }

//----------------------------------------------------------------------
LRESULT CALLBACK WndProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam ) {
    switch (message) {
        case WM_COMMAND: {
			int wmId    = LOWORD(wParam); 
			int wmEvent = HIWORD(wParam); 
			// Parse the menu selections:
            switch (wmId) {
                case IDM_ABOUT:
                    DialogBox( hAppInstance, MAKEINTRESOURCE(IDD_ABOUTBOX), hAppWnd, AboutDialogBox);
                    break;
                case IDM_RUNALLTESTS:
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runAllTestsThread );
                        }
                    break;
                case IDM_TESTS_EXT_DDRAW:
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runExtendedDirectDraw );
                        }
                    break;
                case IDM_TESTS_EXT_D3DKMI:
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runExtendedD3DKMT );
                        }
                    break;
                case IDM_TESTS_EXT_DWM:
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runExtendedCompositionTimingInfo );
                        }
                    break;
                case IDM_TESTS_EXT_SC:
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runExtendedStanislavChromiumTest );
                        }
                    break;


                case IDM_TESTS_EXT_VSYNCSERVER:
                    /*
                    if (!bRunningTests) {
                        ++nNumTestsRun;
                        newThread( runExtendedVsyncServiceTest );
                        }
                    */
                    break;

                case IDM_SETTINGS_REPAINTONMOUSEMOVE:
                    dout( "IDM_SETTINGS_REPAINTONMOUSEMOVE" );
			
                    CheckMenuItem( hMainMenu, IDM_SETTINGS_REPAINTONMOUSEMOVE, isRepaintOnMouseMove()?MF_UNCHECKED:MF_CHECKED );
                    break;

                case IDM_EXIT:
				   DestroyWindow(hWnd);
				   break;
				default:
				   return DefWindowProc(hWnd, message, wParam, lParam);
                }
			break;
            }

        
        case WM_MOUSEMOVE: {
            _m2t = _m1t;
            _m2x = _m1x;
            _m2y = _m1y;

            _m1t = tick();
            _m1x = GET_X_LPARAM(lParam); 
            _m1y = GET_Y_LPARAM(lParam);
            if (isRepaintOnMouseMove()) {
                repaintApp();
                }
            break;
            }

 
        case WM_LBUTTONDBLCLK:
            newThread( slideCursorThread );
            break;

        
        //case WM_ERASEBKGND:
        //    break;
        case WM_PAINT:
            doPaintWindow(hWnd);
			break;
		case WM_DESTROY:
            dout( "WM_DESTROY" );
			PostQuitMessage(0);
			break;
        case WM_USER+7:{
            DWORD t2 = (DWORD)tick();
            DWORD t1 = (DWORD)lParam;
            dout( "%d\t%.1f", (int)wParam, tick2us(t2-t1) );
            break;
            }
        default: {
			return DefWindowProc(hWnd, message, wParam, lParam);
            }
        }
    return 0;
    }

//----------------------------------------------------------------------
ATOM MyRegisterClass( HINSTANCE hInstance ) {
	WNDCLASSEX wcex;
	wcex.cbSize         = sizeof(WNDCLASSEX); 
	wcex.style          = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wcex.lpfnWndProc	= (WNDPROC)WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, (LPCTSTR)IDI_DWM);
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= (LPCSTR)IDC_DWM;
	wcex.lpszClassName	= lpWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, (LPCTSTR)IDI_SMALL);
	return RegisterClassEx(&wcex);
    }
//----------------------------------------------------------------------
BOOL InitInstance( HINSTANCE hInstance, int nCmdShow ) {
    hAppInstance = hInstance;
    hAppWnd = CreateWindow(lpWindowClass, "VSYNC/DWM/D3DKMT Tests -- See crbug.com/467617", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInstance, NULL);
    if (!hAppWnd) {
        return FALSE;
        }
    hMainMenu = GetMenu(hAppWnd);
    hRedBrush = CreateSolidBrush( RGB(255,128,128) );
    hCyanBrush = CreateSolidBrush( RGB(128,255,255) );

    ShowWindow(hAppWnd, nCmdShow);
    UpdateWindow(hAppWnd);
    return TRUE;
    }


//----------------------------------------------------------------------
void ShowOutputInTempFile() {
    char szFile[500];
    if (GetTempPath(sizeof(szFile),szFile)) {
        FILETIME ft;
        GetSystemTimeAsFileTime(&ft);
        sprintf( szFile+lstrlen(szFile),  "vs%ld%ld.txt", szFile, ft.dwLowDateTime, ft.dwHighDateTime );
        HANDLE hFile = CreateFile( szFile, GENERIC_WRITE, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
        LPSTR lpOutput = dout("");
        DWORD dw;
        WriteFile(hFile, lpOutput, lstrlen(lpOutput), &dw, NULL);
        CloseHandle(hFile);
        ShellExecute(GetDesktopWindow(), "open", szFile, NULL, NULL, SW_SHOWMAXIMIZED );
        Sleep(1000);  // hack to leave no file behind.  Notepad is OK with file being deleted after opening the file
        DeleteFile(szFile);
        }
    }
//----------------------------------------------------------------------
void outputWinVersionInfo() {
    DWORD dwVersion = GetVersion();
    DWORD dwMajor = 0xFF & dwVersion;
    DWORD dwMinor = 0xFF & (dwVersion>>8);
    DWORD dwBuild = (dwVersion<0x80000000) ? HIWORD(dwVersion) : 0;
    dout( "Windows Version %d.%d build %d", dwMajor, dwMinor, dwBuild );
    }
//----------------------------------------------------------------------
void outputDateTime() {
    time_t curtime = time(NULL);
    char *pTime = asctime(localtime(&curtime));
    pTime[lstrlen(pTime)-1] = '\0';   // don't want NL
    dout( "%s", pTime );
    }

//----------------------------------------------------------------------
int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow ) {
    setmode(fileno(stdout),O_BINARY);
    tickInit();

    dout( "========== Startup ==========" );
    dout( "dwm-vsync-tests.exe build: %s %s", __DATE__, __TIME__ );
    outputDateTime();
    outputWinVersionInfo();
    dout( "QueryPerformanceFrequency=%I64d", _freq );
    dout( "QueryPerformanceCounter=%I64d", _first );
    dout( "" );

    //getDwmTimingInfo(NULL, NULL, false); // BUGBUG: force load in this thread
    vsyncStartService();

    MyRegisterClass(hInstance);
    if (!InitInstance (hInstance, nCmdShow))  {
		return FALSE;
        }
    HACCEL hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)IDC_DWM);
    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0))  {
        if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg))  {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
            }
        }
    
    // give a running test time to abort
    bAppRunning = false;
    Sleep(50);

    vsyncStopService();

    dout( "Exiting..." );
    Sleep(10);

    if (nNumTestsRun) {
        ShowOutputInTempFile();
        }
    printf( "%s", dout("") );

    return msg.wParam;
    }




//======================================================================
//======================================================================
//======================================================================
//
//  VSYNC synchronization service (SINGLE instance only)
//
//======================================================================
//======================================================================
//======================================================================

static bool _bVsyncServiceRunning=false;
static CRITICAL_SECTION _vilock;
static HANDLE _hVFront = NULL;
static HANDLE _hVBack = NULL;
static HANDLE _hVThread=NULL;

// last vblank time, and period stats
#define PMAX 60
static OSTICK _myvsync=0;
static int _pnums[PMAX];
static int _psum=0;
static int _pnum=0;
static int _poff=0;
static int _scanline=0;

// debug / self monitoring performance stats
static char _vdebug[100]={0};
static OSTICK _vtfirst=0;
static OSTICK _vused=0;
static OSTICK _vsleep=0;



//----------------------------------------------------------------------
void vsyncSuspend( bool bSuspend ) {
    g_bSuspendVsyncServer = bSuspend;
    Sleep(100);
    }
//----------------------------------------------------------------------
LPSTR vsyncGetDebugMessage( _Out_ LPSTR lpMsg ) {
    EnterCriticalSection(&_vilock); 
    lstrcpy( lpMsg, _vdebug ); 
    LeaveCriticalSection(&_vilock);
    return lpMsg;
    }
//----------------------------------------------------------------------
bool vsyncGetNewTimingInfo( _Out_ OSTICK*pvblank, _Out_ OSTICK*pperiod ) {
    bool bRet=false;
    EnterCriticalSection(&_vilock); 
    int period = _pnum ? _psum/_pnum : 0;
    if (pvblank) {
        *pvblank = _myvsync;
        }
    if (pperiod) {
        *pperiod = period;
        }
    bool bOK = _myvsync && period;
    //todo: return default 60Hz info on failure???
    LeaveCriticalSection(&_vilock);
    return bOK;
    }
//----------------------------------------------------------------------
void vsyncHaveNewTimingInfo( _In_ OSTICK now ) {
    // TODO: These numbers are samples, and have micro-jitter.  It IS possible to make
    // a super accurate measurement (see the Hz object at www.vsynctester.com).  That will
    // not help this code to 'wake up' any more accurately, but it might be very useful
    // from a 'monitoring' standpoint.
    EnterCriticalSection(&_vilock);
    if (_myvsync) {
        _psum -= (_pnum==PMAX) ? _pnums[_poff] : 0;
        _pnums[_poff] = (int)(now-_myvsync);
        _psum += _pnums[_poff];
        _poff = (_poff+1)%PMAX;
        _pnum += (_pnum<PMAX)?1:0;
        }
    _myvsync = now;
    LeaveCriticalSection(&_vilock);
    }
//----------------------------------------------------------------------
void vsyncClearTimingInfo() {
    EnterCriticalSection(&_vilock);
    _psum = 0;
    _pnum = 0;
    _poff = 0;
    _myvsync = 0;
    LeaveCriticalSection(&_vilock);
    }
//----------------------------------------------------------------------
OSTICK vsyncGetLast() {
    return _myvsync;
    }


//----------------------------------------------------------------------
// todo: simplify if there is only ONE known waiter
void vsyncSignalAllWaiters() {
    ResetEvent( _hVBack );
    HANDLE temp = _hVFront;
    _hVFront = _hVBack;
    _hVBack = temp;
    SetEvent( _hVBack );
    }

//----------------------------------------------------------------------
void vsyncDoDebug( _In_ LPMONITORINFOEX lpMI ) {
    OSTICK period;
    if (lpMI && vsyncGetNewTimingInfo(NULL, &period)) {
        OSTICK delta = tick()-_vtfirst-_vsleep;
        double mspersec = delta ? tick2ms(_vused-_vsleep)/(tick2ms(delta)/1000) : 0;
        double ms = tick2ms(period);
        double hz = 1000/ms;

        //todo: add dwm timing for comparison???
        char buffer[500];
        LPSTR lpPrimary = lpMI->dwFlags&MONITORINFOF_PRIMARY ? " [PRIMARY]" : "";
        int j = sprintf( buffer, "vsync.server: %s%s %.3fms %.3fHz (using %.3f ms/sec) DWM=%d", lpMI->szDevice, lpPrimary, ms, hz, mspersec, isDwmEnabled() );

        EnterCriticalSection(&_vilock); 
        CopyMemory( _vdebug, buffer, j+1 );
        LeaveCriticalSection(&_vilock);
        repaintApp();
        }
    }


//----------------------------------------------------------------------
bool vsyncGetAppMonitorInfo( _Out_ LPMONITORINFOEX lpMI ) {
    bool bOK=false;
    if (hAppWnd) {
        HMONITOR hMonitor = myMonitorFromWindow(hAppWnd, MONITOR_DEFAULTTONEAREST);
        lpMI->cbSize = sizeof(*lpMI);
        bOK =  hMonitor && myGetMonitorInfo(hMonitor, lpMI);
        }
    return bOK;
    }
//----------------------------------------------------------------------
//LPSTR getVsyncToDisplayDevice( _Out_ LPSTR lpDevice ) {
//    lpDevice[0] = '\0';
//    bool bDwmActive = isDwmEnabled(); ///getDwmTimingInfo(NULL, NULL, false);
//    // TEST: FORCE TO NEAREST
//    bDwmActive = false;
//    // if dwm *IS* active, sync only to 'primary' display, regardless of what display app is on
//    if (bDwmActive) {
//        DISPLAY_DEVICE dd;
//        dd.cb = sizeof(dd);
//        for (int loop=0; myEnumDisplayDevices(NULL, loop, &dd, 0); ++loop) {
//            if (dd.StateFlags&DISPLAY_DEVICE_PRIMARY_DEVICE) {
//                lstrcpy( lpDevice, (LPSTR)dd.DeviceName );
//                break;
//                }
//            }
//        }
//    // if dwm is *NOT* active (or on a dwm failure above) find monitor that app window is on
//    if (!*lpDevice && hAppWnd) {
//        HMONITOR hMonitor = myMonitorFromWindow(hAppWnd, MONITOR_DEFAULTTONEAREST);
//        MONITORINFOEX mi;
//        mi.cbSize = sizeof(mi);
//        if (hMonitor && myGetMonitorInfo(hMonitor, &mi)) {
//            lstrcpy( lpDevice, mi.szDevice );
//            }
//        }
//    return lpDevice;
//    }





//----------------------------------------------------------------------
// close prior display adapter (if any)
void vsyncClosePreviousAdapter( _InOut_ D3DKMT_WAITFORVERTICALBLANKEVENT*pwe ) {
    if (pwe->hAdapter) {
        D3DKMT_CLOSEADAPTER ca;
        ca.hAdapter = pwe->hAdapter;
        NTSTATUS res = myD3DKMTCloseAdapter(&ca);
        dout( "vsync.close.adapter=%d", res );
        pwe->hAdapter = 0;
        }
    }
//----------------------------------------------------------------------
// open display adapter
bool vsyncFindAndOpenAdapter( _Out_ D3DKMT_WAITFORVERTICALBLANKEVENT*pwe, _Out_  LPMONITORINFOEX lpMI ) {
    bool bBound=false;            
    D3DKMT_OPENADAPTERFROMHDC oa;

    bool bHaveMonitor = vsyncGetAppMonitorInfo(lpMI);
    if (bHaveMonitor) {
        oa.hDc = CreateDC(NULL, lpMI->szDevice, NULL, NULL);
        dout( "vsync.bind device='%s' hDC=%d", lpMI->szDevice, oa.hDc );
        if (oa.hDc) {
            if (STATUS_SUCCESS==myD3DKMTOpenAdapterFromHdc(&oa)) {
                pwe->hAdapter = oa.hAdapter;
                pwe->hDevice = 0;
                pwe->VidPnSourceId = oa.VidPnSourceId;
                bBound = true;
                dout( "vsync.bound hAdapter=%d", oa.hAdapter );
                }
            DeleteDC( oa.hDc );
            }
        }
    return bBound;
    }
//----------------------------------------------------------------------
void vsleep( DWORD ms ) {
    OSTICK t1 = tick();
    Sleep(ms);
    OSTICK t2 = tick();
    _vsleep += (t2-t1);
    }

//----------------------------------------------------------------------
// BUGBUG: work around for DirectDraw conflict, when that test is running
void vsyncImplementSuspend() {
    if (g_bSuspendVsyncServer) {
        dout( "vsync.suspended" );
        while (_bVsyncServiceRunning && g_bSuspendVsyncServer) {
            vsleep(10);
            }
        dout( "vsync.running" );
        }
    }
//----------------------------------------------------------------------
// every so often, confirm we are still syncing to the proper display device
bool vsyncValidateConfiguration( _In_ LPSTR lpDevice, _In_ OSTICK now ) {
    bool bBound = true;
    static OSTICK last=0;
    if (tick2ms(now-last)>500) {
        last = now;
        MONITORINFOEX mi;
        bool bHaveMonitor = vsyncGetAppMonitorInfo(&mi);
        vsyncDoDebug(bHaveMonitor?&mi:NULL);
        bBound = bHaveMonitor && (0==strcmp(lpDevice,mi.szDevice));
        if (!bBound) {
            dout( "vsync.validate.configuration rebind needed" );
            }
        }
    return bBound;
    }
//----------------------------------------------------------------------
OSTICK vsyncAddPerformanceStats( OSTICK now, OSTICK cpu ) {
    static int nTimes=0;
    _vused += cpu;
    if (nTimes<10) {
        _vtfirst = now;
        _vsleep = 0;
        _vused = 0;
        }
    ++nTimes;
    return now;
    }

//----------------------------------------------------------------------
// a hack, because this relies up global vars/handles
D3DKMT_GETSCANLINE g_gsl={0,0,0,0};
int vsyncGetScanLine() {
    D3DKMT_GETSCANLINE gsl = g_gsl;
    bool bOK = gsl.hAdapter ? (STATUS_SUCCESS==myD3DKMTGetScanLine(&gsl)) : false;

    if (bOK) {
        D3DKMT_DISPLAYMODE dm;
        if (getCurrentDisplayMode(g_gsl.hAdapter, g_gsl.VidPnSourceId, &dm)) {
            ///dout( "%d %d" );
            ///-- %d", dm.Width, dm.Height, dm.ScanLineOrdering );
            }
        }

    // return value: 0=error; otherwise scan line; if negative, then in vblank
    return bOK ? (int)max(1,gsl.ScanLine) * (gsl.InVerticalBlank?-1:1) : 0;
    }
int vsyncGSL() {
    return _scanline;
    }

//----------------------------------------------------------------------
// DWM ON  -- syncs to the 'primary' display Hz, and follows changes in primary display
// DWM OFF -- syncs to the display Hz that this app window is (mostly) on
DWORD THREADPROC vsyncThread(LPVOID lpParam) {
    dout( "+vsyncThread()" );

    // When D3DKMT wakes up this thread, THIS CODE MUST RUN ASAP -- it runs in no time and then IMMEDIATELY right back to waiting/sleeping
    SetThreadPriority( GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL );

    // perf mon
    OSTICK t1 = vsyncAddPerformanceStats(tick(), 0);
    
    D3DKMT_WAITFORVERTICALBLANKEVENT we={0,0,0}; // zero hAdapter; hDevice, VidPnSourceId 
    MONITORINFOEX mi;
    mi.szDevice[0] = '\0';

    bool bBound = false;
    while (_bVsyncServiceRunning) {
        vsyncImplementSuspend();  // only needed due to DirectDraw test code far above; otherwise remove

        // if not bound to display device, reset and try to bind to the display device
        if (!bBound) {
            vsyncClearTimingInfo();
            vsyncClosePreviousAdapter(&we);
            bBound = vsyncFindAndOpenAdapter(&we, &mi);
            dout( "-vsyncRebind=%d", bBound );
            if (!bBound) {
                dout( "vsync.failsafe.wait" );
                vsleep(200);  // should never get here (adapter fails to open), so a failsafe
                }
            else {
                D3DKMT_DISPLAYMODE dm;
                if (getCurrentDisplayMode(we.hAdapter, we.VidPnSourceId, &dm)) {
                    dout( "VSYNC INFO: size = %dx%d  hz=%d/%d", dm.Width, dm.Height, dm.RefreshRate.Numerator, dm.RefreshRate.Denominator );
                    }
                }
            }

        g_gsl.hAdapter = 0;

        // if bound to display device...
        if (bBound) {

            g_gsl.hAdapter = we.hAdapter;
            g_gsl.VidPnSourceId = we.VidPnSourceId;


            // wait for vblank on display device
            OSTICK t2 = tick();
            OSTICK cpu = t2-t1;
            bool bWaited = (STATUS_SUCCESS==myD3DKMTWaitForVerticalBlankEvent(&we));
            t1 = tick();

            vsyncAddPerformanceStats(t1, cpu);

            if (bWaited) {
                _scanline = vsyncGetScanLine();
                ++g_nFrame;
                vsyncHaveNewTimingInfo( t1 );
                vsyncSignalAllWaiters();
                bBound = vsyncValidateConfiguration(mi.szDevice, t1);   // and debug string
                if (!bRunningTests) {
                    repaintApp();
                    }
                }
            else {
                // TODO: This happens in Win 8.1 when the display is 'turned off'
                dout( "vsync.WaitForVerticalBlankEvent FAILED" );
                vsleep(500);
                bBound = false;
                }
            }
        }

    dout( "-vsyncThread()" );
    _hVBack = NULL;
    _hVFront = NULL;
    _scanline = 0;
    return 0;
    }


// NOTE: waking up on an external signal is OPTIONAL
// or another way: per thread (TLS) wait object, on a wait, into queue; on notify, all in queue.
//----------------------------------------------------------------------
// Designed to that any number of callers can use at once, and wake up on VSYNC
// reason for return: vsync interrupted (true) or timeout (false)
bool vsyncSleepWithVsyncInterrupt( _In_ DWORD dwMilliseconds ) {
    return WAIT_OBJECT_0==WaitForSingleObject(_hVFront, dwMilliseconds);
    }

//----------------------------------------------------------------------
// not multi-thread safe (intended to be called at app startup / WinMain)
void vsyncStartService() {
    if (!_bVsyncServiceRunning) {
        _bVsyncServiceRunning = true;
        InitializeCriticalSectionAndSpinCount(&_vilock,8);  // protects such a small section of code, spining will always be much faster than actual waiting in the OS
        _hVFront = CreateEvent(NULL, TRUE, FALSE, NULL);
        _hVBack = CreateEvent(NULL, TRUE, FALSE, NULL);
        _hVThread = newThread( vsyncThread );
        }
    }
//----------------------------------------------------------------------
// not multi-thread safe (intended to be called at app shutdown / WinMain)
void vsyncStopService() {
    if (_bVsyncServiceRunning) {
        _bVsyncServiceRunning = false;
        WaitForSingleObject(_hVThread,5000);
        }
    }



